// decodable-nnet-online-for-hark.cc
// If you want to put this file under the directory, it file is placed in under the "nnet" directory.

#ifdef _MSC_VER
#define NOMINMAX
#endif
#include "decodable-nnet-online-for-hark.h"

//#include "nnet/online-nnet-decodable.h"
//#include "nnet2/online-nnet2-decodable.h"
//#include <nnet3/online-nnet3-decodable-simple.h>
#include "nnet3/nnet-utils.h"

namespace kaldi {
namespace nnet1 {

DecodableNnetOnlineForHark::DecodableNnetOnlineForHark(
        const kaldi::nnet1::Nnet &nnet,
        kaldi::nnet1::PdfPrior &pdf_prior,
        const kaldi::TransitionModel &trans_model,
        const DecodableNnetOnlineForHarkOptions &opts,
        kaldi::OnlineFeatureInterface *features_if):
    features_if_(features_if),
    nnet_(nnet),
    pdf_prior_(pdf_prior),
    trans_model_(trans_model),
    opts_(opts),
    feat_dim_(features_if->Dim()),
    num_pdfs_(nnet.OutputDim()),
    begin_frame_(-1) {
  KALDI_ASSERT(opts_.max_nnet_batch_size > 0);
}

kaldi::BaseFloat DecodableNnetOnlineForHark::LogLikelihood(kaldi::int32 frame, kaldi::int32 index) {
  ComputeForFrame(frame);
  kaldi::int32 pdf_id = trans_model_.TransitionIdToPdf(index);
  KALDI_ASSERT(frame >= begin_frame_ && (frame < begin_frame_ + scaled_loglikes_.NumRows()));
  return scaled_loglikes_(frame - begin_frame_, pdf_id);
}

bool DecodableNnetOnlineForHark::IsLastFrame(kaldi::int32 frame) const {
  return features_if_->IsLastFrame(frame);
}

kaldi::int32 DecodableNnetOnlineForHark::NumFramesReady() const {
  return features_if_->NumFramesReady();
}

void DecodableNnetOnlineForHark::ComputeForFrame(kaldi::int32 frame) {
  kaldi::int32 features_ready = features_if_->NumFramesReady();
  //bool input_finished = features_if_->IsLastFrame(features_ready - 1);
  //KALDI_LOG << features_ready;
  if(frame >= begin_frame_ && (frame < begin_frame_ + scaled_loglikes_.NumRows())){
    return;
  }

  kaldi::int32 input_frame_begin = frame;
  kaldi::int32 max_possible_input_frame_end = features_ready;
  kaldi::int32 input_frame_end =
      std::min<kaldi::int32>(max_possible_input_frame_end,
                             input_frame_begin + opts_.max_nnet_batch_size);

  KALDI_ASSERT(input_frame_end > input_frame_begin);
  kaldi::Matrix<kaldi::BaseFloat> features(input_frame_end - input_frame_begin, feat_dim_);
  for(kaldi::int32 t = input_frame_begin; t < input_frame_end; t++){
    kaldi::SubVector<kaldi::BaseFloat> row(features, t - input_frame_begin);
    features_if_->GetFrame(t, &row);
  }
  kaldi::CuMatrix<kaldi::BaseFloat> cu_features;
  cu_features.Swap(&features);

  kaldi::int32 num_frames_out = input_frame_end - input_frame_begin;
  kaldi::CuMatrix<kaldi::BaseFloat> cu_posteriors(num_frames_out, num_pdfs_);

  nnet_.Feedforward(cu_features, &cu_posteriors);
  pdf_prior_.SubtractOnLogpost(&cu_posteriors);
  cu_posteriors.Scale(opts_.acoustic_scale);

  scaled_loglikes_.Resize(0, 0);
  cu_posteriors.Swap(&scaled_loglikes_);

  begin_frame_ = frame;
}

} // namespace nnet1
////////////////////////////////////////////////////////////////////////////////
namespace nnet2 {

DecodableNnet2OnlineForHark::DecodableNnet2OnlineForHark(
        const kaldi::nnet2::AmNnet &nnet,
        const kaldi::TransitionModel &trans_model,
        const DecodableNnet2OnlineForHarkOptions &opts,
        kaldi::OnlineFeatureInterface *features_if):
    features_if_(features_if),
    nnet_(nnet),
    trans_model_(trans_model),
    opts_(opts),
    feat_dim_(features_if->Dim()),
    left_context_(nnet.GetNnet().LeftContext()),
    right_context_(nnet.GetNnet().RightContext()),
    num_pdfs_(nnet.GetNnet().OutputDim()),
    begin_frame_(-1) {
  KALDI_ASSERT(opts_.max_nnet_batch_size > 0);
  log_priors_ = nnet_.Priors();
  KALDI_ASSERT(log_priors_.Dim() == trans_model_.NumPdfs() &&
               "Priors in neural network not set up (or mismatch with transition model).");
  log_priors_.ApplyLog();
}

kaldi::BaseFloat DecodableNnet2OnlineForHark::LogLikelihood(kaldi::int32 frame, kaldi::int32 index) {
  ComputeForFrame(frame);
  kaldi::int32 pdf_id = trans_model_.TransitionIdToPdf(index);
  KALDI_ASSERT(frame >= begin_frame_ && (frame < begin_frame_ + scaled_loglikes_.NumRows()));
  return scaled_loglikes_(frame - begin_frame_, pdf_id);
}

bool DecodableNnet2OnlineForHark::IsLastFrame(kaldi::int32 frame) const {
  //opts_.pad_input == true is normal case
  return features_if_->IsLastFrame(frame + (opts_.pad_input ? 0 : left_context_ + right_context_));
}

kaldi::int32 DecodableNnet2OnlineForHark::NumFramesReady() const {
  kaldi::int32 features_ready = features_if_->NumFramesReady();
  if(features_ready == 0){ return 0; }
  bool input_finished = features_if_->IsLastFrame(features_ready - 1);
  if(opts_.pad_input){
    // normal case... we'll pad with duplicates of first + last frame to get the
    // required left and right context.
    if(input_finished){ return features_ready; }
    else{ return std::max<kaldi::int32>(0, features_ready - right_context_); }
  }
  else{
    return std::max<kaldi::int32>(0, features_ready - right_context_ - left_context_);
  }
}

void DecodableNnet2OnlineForHark::ComputeForFrame(kaldi::int32 frame) {
  kaldi::int32 features_ready = features_if_->NumFramesReady();
  bool input_finished = features_if_->IsLastFrame(features_ready - 1);  
  KALDI_ASSERT(frame >= 0);
  if(frame >= begin_frame_ && (frame < begin_frame_ + scaled_loglikes_.NumRows())){
    return;
  }
  KALDI_ASSERT(frame < NumFramesReady());

  kaldi::int32 input_frame_begin = frame - (opts_.pad_input ? left_context_ : 0);
  kaldi::int32 max_possible_input_frame_end = features_ready + ((input_finished && opts_.pad_input) ? right_context_ : 0);
  kaldi::int32 input_frame_end =
      std::min<kaldi::int32>(max_possible_input_frame_end,
                             input_frame_begin + left_context_ + right_context_ + opts_.max_nnet_batch_size);

  KALDI_ASSERT(input_frame_end > input_frame_begin);
  kaldi::Matrix<kaldi::BaseFloat> features(input_frame_end - input_frame_begin, feat_dim_);
  for(kaldi::int32 t = input_frame_begin; t < input_frame_end; t++){
    kaldi::SubVector<kaldi::BaseFloat> row(features, t - input_frame_begin);
    kaldi::int32 t_modified = (t < 0 ? 0 : (t >= features_ready ? features_ready - 1 : t)); // This statements take care of "pad_input"
    features_if_->GetFrame(t_modified, &row);
  }
  kaldi::CuMatrix<kaldi::BaseFloat> cu_features; 
  cu_features.Swap(&features);  // Copy to GPU, if we're using one.

  kaldi::int32 num_frames_out = input_frame_end - input_frame_begin - left_context_ - right_context_;
  kaldi::CuMatrix<kaldi::BaseFloat> cu_posteriors(num_frames_out, num_pdfs_);

  // The "false" below tells it not to pad the input: we've already done
  // any padding that we needed to do.
  kaldi::nnet2::NnetComputation(nnet_.GetNnet(), cu_features, false, &cu_posteriors);
  
  cu_posteriors.ApplyFloor(1.0e-20); // Avoid log of zero which leads to NaN.
  cu_posteriors.ApplyLog();
  // subtract log-prior (divide by prior)
  cu_posteriors.AddVecToRows(-1.0, log_priors_);
  // apply probability scale.
  cu_posteriors.Scale(opts_.acoustic_scale);

  // Transfer the scores the CPU for faster access by the decoding process.
  scaled_loglikes_.Resize(0, 0);
  cu_posteriors.Swap(&scaled_loglikes_);

  begin_frame_ = frame;
}


} // namespace nnet2
////////////////////////////////////////////////////////////////////////////////
namespace nnet3 {

DecodableNnet3SimpleOnlineForHark::DecodableNnet3SimpleOnlineForHark(
        const kaldi::nnet3::AmNnetSimple &nnet,
        const kaldi::TransitionModel &trans_model,
        const DecodableNnet3OnlineForHarkOptions &opts,
        kaldi::OnlineFeatureInterface *features_if):
    compiler_(nnet.GetNnet(), opts.optimize_config),
    features_if_(features_if),
    nnet_(nnet),
    trans_model_(trans_model),
    opts_(opts),
    feat_dim_(features_if->Dim()),
    num_pdfs_(nnet.GetNnet().OutputDim("output")),
    begin_frame_(-1) {
  KALDI_ASSERT(opts_.max_nnet_batch_size > 0);
  log_priors_ = nnet_.Priors();
  KALDI_ASSERT((log_priors_.Dim() == 0 || log_priors_.Dim() == trans_model_.NumPdfs()) &&
               "Priors in neural network must match with transition model (if exist).");

  ComputeSimpleNnetContext(nnet_.GetNnet(), &left_context_, &right_context_);
  log_priors_.ApplyLog();

  kaldi::int32 input_dim = nnet_.GetNnet().InputDim("input");
  kaldi::int32 ivector_dim = std::max<kaldi::int32>(0, nnet_.GetNnet().InputDim("ivector"));
  kaldi::int32 feature_dim = features_if_->Dim();
  if(feature_dim != input_dim + ivector_dim){
    KALDI_ERR << "Dimension of features " << feature_dim << " does not equal "
              << "input dim " << input_dim << " + ivector dim " << ivector_dim
              << " of neural network.  Likely the config and neural net mismatch.";
  }
}

kaldi::BaseFloat DecodableNnet3SimpleOnlineForHark::LogLikelihood(kaldi::int32 frame, kaldi::int32 index) {
  ComputeForFrame(frame);
  kaldi::int32 pdf_id = trans_model_.TransitionIdToPdf(index);
  KALDI_ASSERT(frame >= begin_frame_ && (frame < begin_frame_ + scaled_loglikes_.NumRows()));
  return scaled_loglikes_(frame - begin_frame_, pdf_id);
}

bool DecodableNnet3SimpleOnlineForHark::IsLastFrame(kaldi::int32 frame) const {
  KALDI_ASSERT(false && "Method is not imlemented");
  return false;
}

kaldi::int32 DecodableNnet3SimpleOnlineForHark::NumFramesReady() const {
  kaldi::int32 features_ready = features_if_->NumFramesReady();
  if(features_ready == 0){ return 0; }
  bool input_finished = features_if_->IsLastFrame(features_ready - 1);
  if(opts_.pad_input){
    // normal case... we'll pad with duplicates of first + last frame to get the
    // required left and right context.
    if(input_finished){ return NumSubsampledFrames(features_ready); }
    else{ return std::max<kaldi::int32>(0, NumSubsampledFrames(features_ready - right_context_)); }
  }
  else{
    return std::max<kaldi::int32>(0, NumSubsampledFrames(features_ready - right_context_ - left_context_));
  }
}

kaldi::int32 DecodableNnet3SimpleOnlineForHark::NumSubsampledFrames(kaldi::int32 num_frames) const {
  return (num_frames) / opts_.frame_subsampling_factor;
}

void DecodableNnet3SimpleOnlineForHark::ComputeForFrame(kaldi::int32 subsampled_frame) {
  kaldi::int32 features_ready = features_if_->NumFramesReady();
  bool input_finished = features_if_->IsLastFrame(features_ready - 1);
  KALDI_ASSERT(subsampled_frame >= 0);
  if(subsampled_frame >= begin_frame_ && (subsampled_frame < begin_frame_ + scaled_loglikes_.NumRows())){
    return;
  }
  KALDI_ASSERT(subsampled_frame < NumFramesReady());

  kaldi::int32 subsample = opts_.frame_subsampling_factor;

  kaldi::int32 input_frame_begin = subsampled_frame * subsample  - (opts_.pad_input ? left_context_ : 0);
  kaldi::int32 max_possible_input_frame_end = features_ready + ((input_finished && opts_.pad_input) ? right_context_ : 0);
  kaldi::int32 input_frame_end =
      std::min<kaldi::int32>(max_possible_input_frame_end,
                             input_frame_begin + left_context_ + right_context_ + opts_.max_nnet_batch_size);

  KALDI_ASSERT(input_frame_end > input_frame_begin);
  kaldi::Matrix<kaldi::BaseFloat> features(input_frame_end - input_frame_begin, feat_dim_);
  for(kaldi::int32 t = input_frame_begin; t < input_frame_end; t++){
    kaldi::SubVector<kaldi::BaseFloat> row(features, t - input_frame_begin);
    kaldi::int32 t_modified = (t < 0 ? 0 : (t >= features_ready ? features_ready - 1 : t)); // This statements take care of "pad_input"
    features_if_->GetFrame(t_modified, &row);
  }

  kaldi::int32 num_subsampled_frames =
      NumSubsampledFrames(input_frame_end - input_frame_begin - left_context_ - right_context_);
  kaldi::int32 input_feat_dim = nnet_.GetNnet().InputDim("input");
  kaldi::int32 ivector_dim = nnet_.GetNnet().InputDim("ivector");
  // MFCCs in the left chunk
  kaldi::SubMatrix<kaldi::BaseFloat> input_feat_mat(features.ColRange(0, input_feat_dim));

  kaldi::Vector<kaldi::BaseFloat> input_ivector;
  if(ivector_dim != -1){
    // iVectors in the right chunk
    KALDI_ASSERT(features.NumCols() == input_feat_dim + ivector_dim && "Mismatch in features dim");
    kaldi::SubMatrix<kaldi::BaseFloat> ivector_mat(features.ColRange(input_feat_dim, ivector_dim));
    // Get last ivector... not sure if GetCurrentIvector is needed in the online context
    // I think it should work fine just getting the last row for testing
    input_ivector = ivector_mat.Row(ivector_mat.NumRows() - 1);
  }

  DoNnetComputation(input_frame_begin,
                    input_feat_mat, input_ivector,
                    subsampled_frame * subsample,
                    num_subsampled_frames);

  begin_frame_ = subsampled_frame;
}

void DecodableNnet3SimpleOnlineForHark::DoNnetComputation(
    kaldi::int32 input_t_start,
    const kaldi::MatrixBase<kaldi::BaseFloat> &input_feats,
    const kaldi::VectorBase<kaldi::BaseFloat> &ivector,
    kaldi::int32 output_t_start,
    kaldi::int32 num_subsampled_frames) {
  ComputationRequest request;
  request.need_model_derivative = false;
  request.store_component_stats = false;

  bool shift_time = true; // shift the 'input' and 'output' to a consistent
                          // time, to take advantage of caching in the compiler.
                          // An optimization.
  kaldi::int32 time_offset = (shift_time ? -output_t_start : 0);

  // First add the regular features-- named "input".
  request.inputs.reserve(2);
  request.inputs.push_back(
      kaldi::nnet3::IoSpecification("input", time_offset + input_t_start, time_offset + input_t_start + input_feats.NumRows())
  );
  if(ivector.Dim() != 0){
    std::vector<kaldi::nnet3::Index> indexes;
    indexes.push_back(kaldi::nnet3::Index(0, 0, 0));
    request.inputs.push_back(IoSpecification("ivector", indexes));
  }
  kaldi::nnet3::IoSpecification output_spec;
  output_spec.name = "output";
  output_spec.has_deriv = false;
  kaldi::int32 subsample = opts_.frame_subsampling_factor;
  output_spec.indexes.resize(num_subsampled_frames);
  // leave n and x values at 0 (the constructor sets these).
  for(kaldi::int32 i = 0; i < num_subsampled_frames; i++){
    output_spec.indexes[i].t = time_offset + output_t_start + i * subsample;
  }
  request.outputs.resize(1);
  request.outputs[0].Swap(&output_spec);

  const kaldi::nnet3::NnetComputation *computation = compiler_.Compile(request);
  kaldi::nnet3::Nnet *nnet_to_update = NULL;  // we're not doing any update.
  kaldi::nnet3::NnetComputer computer(opts_.compute_config, *computation, nnet_.GetNnet(), nnet_to_update);

  kaldi::CuMatrix<kaldi::BaseFloat> input_feats_cu(input_feats);
  computer.AcceptInput("input", &input_feats_cu);
  kaldi::CuMatrix<kaldi::BaseFloat> ivector_feats_cu;
  if(ivector.Dim() > 0){
    ivector_feats_cu.Resize(1, ivector.Dim());
    ivector_feats_cu.Row(0).CopyFromVec(ivector);
    computer.AcceptInput("ivector", &ivector_feats_cu);
  }
  computer.Forward();
  kaldi::CuMatrix<kaldi::BaseFloat> cu_output;
  computer.GetOutputDestructive("output", &cu_output);
  if(log_priors_.Dim() != 0){
    cu_output.AddVecToRows(-1.0, log_priors_); // subtract log-prior (divide by prior)
  }
  cu_output.Scale(opts_.acoustic_scale); // apply the acoustic scale
  scaled_loglikes_.Resize(0, 0);
  cu_output.Swap(&scaled_loglikes_); // the following statement just swaps the pointers if we're not using a GPU.
}




} // namespace nnet3
} // namespace kaldi
