// decodable-nnet-online-for-hark.h
// If you want to put this file under the directory, it file is placed in under the "nnet" directory.
// Implements like "nnet2/online-nnet2-decodable.h".
// Implements like "nnet3/online-nnet3-decodable.h".

#ifndef KALDI_DECODABLE_NNET_ONLINE_FOR_HARK_H_
#define KALDI_DECODABLE_NNET_ONLINE_FOR_HARK_H_

#include "itf/online-feature-itf.h"
#include "itf/decodable-itf.h"
#include "hmm/transition-model.h"
// nnet1
#include "nnet/nnet-nnet.h"
#include "nnet/nnet-pdf-prior.h"
// nnet2
#include "nnet2/am-nnet.h"
#include "nnet2/nnet-compute.h"
// nnet3
#include "nnet3/am-nnet-simple.h"
#include "nnet3/nnet-compute.h"
#include "nnet3/nnet-optimize.h"

namespace kaldi {
namespace nnet1 {

struct DecodableNnetOnlineForHarkOptions {
  kaldi::BaseFloat acoustic_scale;
  kaldi::int32 max_nnet_batch_size;

  DecodableNnetOnlineForHarkOptions():
      acoustic_scale(0.1),
      max_nnet_batch_size(128/*256*/) { }

  // for kaldi compatibility
  void Register(OptionsItf *opts) {
    opts->Register("acoustic-scale", &acoustic_scale,
                   "Scaling factor for acoustic likelihoods");
    opts->Register("max-nnet-batch-size", &max_nnet_batch_size,
                   "Maximum batch size we use in neural-network decodable object, "
                   "in cases where we are not constrained by currently available "
                   "frames (this will rarely make a difference)");
  }
  // for avoid of option name conflict in multi nnet type support
  void SetAcousticScale(kaldi::BaseFloat arg) {
    acoustic_scale = arg;
  }
  void SetMaxNnetBatchSize(kaldi::int32 arg) {
    max_nnet_batch_size = arg;
  }
};

class DecodableNnetOnlineForHark : public DecodableInterface {
 public:
  DecodableNnetOnlineForHark(const kaldi::nnet1::Nnet &nnet,
                             kaldi::nnet1::PdfPrior &pdf_prior,
                             const kaldi::TransitionModel &trans_model,
                             const DecodableNnetOnlineForHarkOptions &opts,
//                             kaldi::BaseFloat ac_scale,
                             kaldi::OnlineFeatureInterface *features_if);

  /// Returns the scaled log likelihood
  virtual kaldi::BaseFloat LogLikelihood(kaldi::int32 frame, kaldi::int32 index);

  virtual bool IsLastFrame(kaldi::int32 frame) const;

  virtual kaldi::int32 NumFramesReady() const;

  /// Indices are one-based!  This is for compatibility with OpenFst.
  virtual kaldi::int32 NumIndices() const { return trans_model_.NumTransitionIds(); }

 private:
  void ComputeForFrame(kaldi::int32 frame);

  kaldi::OnlineFeatureInterface *features_if_;
  kaldi::nnet1::Nnet nnet_;
  kaldi::nnet1::PdfPrior &pdf_prior_;  // log-priors taken from the model.
  const kaldi::TransitionModel &trans_model_;
  DecodableNnetOnlineForHarkOptions opts_;
//  kaldi::BaseFloat ac_scale_;
//  kaldi::int32 max_nnet_batch_size_;
  kaldi::int32 feat_dim_;
  kaldi::int32 num_pdfs_;
  kaldi::int32 begin_frame_;

  kaldi::Matrix<kaldi::BaseFloat> scaled_loglikes_;

  //KALDI_DISALLOW_COPY_AND_ASSIGN(DecodableNnetOnline);
};

} // namespace nnet1

namespace nnet2 {

struct DecodableNnet2OnlineForHarkOptions {
  kaldi::BaseFloat acoustic_scale;
  bool pad_input;
  kaldi::int32 max_nnet_batch_size;
  
  DecodableNnet2OnlineForHarkOptions():
      acoustic_scale(0.1),
      pad_input(true),
      max_nnet_batch_size(256) { }

  // for kaldi compatibility
  void Register(OptionsItf *opts) {
    opts->Register("acoustic-scale", &acoustic_scale,
                   "Scaling factor for acoustic likelihoods");
    opts->Register("pad-input", &pad_input,
                   "If true, pad acoustic features with required acoustic context "
                   "past edges of file.");
    opts->Register("max-nnet-batch-size", &max_nnet_batch_size,
                   "Maximum batch size we use in neural-network decodable object, "
                   "in cases where we are not constrained by currently available "
                   "frames (this will rarely make a difference)");
  }
  // for avoid of option name conflict in multi nnet type support
  void SetAcousticScale(kaldi::BaseFloat arg) {
    acoustic_scale = arg;
  }
  void SetPadInput(bool arg) {
    pad_input = arg;
  }
  void SetMaxNnetBatchSize(kaldi::int32 arg) {
    max_nnet_batch_size = arg;
  }
};

class DecodableNnet2OnlineForHark: public DecodableInterface {
 public:
  DecodableNnet2OnlineForHark(const kaldi::nnet2::AmNnet &nnet,
                              const kaldi::TransitionModel &trans_model,
                              const DecodableNnet2OnlineForHarkOptions &opts,
                              kaldi::OnlineFeatureInterface *features_if);

  /// Returns the scaled log likelihood
  virtual kaldi::BaseFloat LogLikelihood(kaldi::int32 frame, kaldi::int32 index);

  virtual bool IsLastFrame(kaldi::int32 frame) const;

  virtual kaldi::int32 NumFramesReady() const;  

  /// Indices are one-based!  This is for compatibility with OpenFst.
  virtual kaldi::int32 NumIndices() const { return trans_model_.NumTransitionIds(); }

 private:
  void ComputeForFrame(kaldi::int32 frame);

  kaldi::OnlineFeatureInterface *features_if_;
  const kaldi::nnet2::AmNnet &nnet_;
  const kaldi::TransitionModel &trans_model_;
  DecodableNnet2OnlineForHarkOptions opts_;
  kaldi::CuVector<kaldi::BaseFloat> log_priors_;
  kaldi::int32 feat_dim_;
  kaldi::int32 left_context_;
  kaldi::int32 right_context_;
  kaldi::int32 num_pdfs_;
  kaldi::int32 begin_frame_;

  kaldi::Matrix<kaldi::BaseFloat> scaled_loglikes_;

  //KALDI_DISALLOW_COPY_AND_ASSIGN(DecodableNnet2OnlineForHark);
};


} // namespace nnet2

namespace nnet3 {

struct DecodableNnet3OnlineForHarkOptions {
  kaldi::int32 frame_subsampling_factor;
  kaldi::BaseFloat acoustic_scale;
  bool pad_input;
  kaldi::int32 max_nnet_batch_size;
  kaldi::nnet3::NnetComputeOptions compute_config;
  kaldi::nnet3::NnetOptimizeOptions optimize_config;

  DecodableNnet3OnlineForHarkOptions():
      frame_subsampling_factor(1),
      acoustic_scale(0.1),
      pad_input(true),
      max_nnet_batch_size(256) { }

  // for kaldi compatibility
  void Register(kaldi::OptionsItf *opts) {
    opts->Register("acoustic-scale", &acoustic_scale,
                   "Scaling factor for acoustic likelihoods");
    opts->Register("pad-input", &pad_input,
                   "If true, pad acoustic features with required acoustic context "
                   "past edges of file.");
    opts->Register("max-nnet-batch-size", &max_nnet_batch_size,
                   "Maximum batch size we use in neural-network decodable object, "
                   "in cases where we are not constrained by currently available "
                   "frames (this will rarely make a difference)");
    opts->Register("frame-subsampling-factor", &frame_subsampling_factor,
                   "Required if the frame-rate of the output (e.g. in 'chain' "
                   "models) is less than the frame-rate of the original "
                   "alignment.");

    // register the optimization options with the prefix "optimization".
    ParseOptions optimization_opts("optimization", opts);
    optimize_config.Register(&optimization_opts);

    // register the compute options with the prefix "computation".
    ParseOptions compute_opts("computation", opts);
    compute_config.Register(&compute_opts);
  }
  // for avoid of option name conflict in multi nnet type support
  void SetFrameSubsamplingFactor(kaldi::int32 arg) {
    frame_subsampling_factor = arg;
  }
  void SetAcousticScale(kaldi::BaseFloat arg) {
    acoustic_scale = arg;
  }
  void SetPadInput(bool arg) {
    pad_input = arg;
  }
  void SetMaxNnetBatchSize(kaldi::int32 arg) {
    max_nnet_batch_size = arg;
  }
  void RegisterOptimizationOptions(kaldi::OptionsItf *opts) {
    // register the optimization options with the prefix "optimization".
    ParseOptions optimization_opts("optimization", opts);
    optimize_config.Register(&optimization_opts);
  }
  void RegisterConputeOptions(kaldi::OptionsItf *opts) {
    // register the compute options with the prefix "computation".
    ParseOptions compute_opts("computation", opts);
    compute_config.Register(&compute_opts);
  }
};

class DecodableNnet3SimpleOnlineForHark: public DecodableInterface {
 public:
  DecodableNnet3SimpleOnlineForHark(const kaldi::nnet3::AmNnetSimple &am_nnet,
                                    const kaldi::TransitionModel &trans_model,
                                    const DecodableNnet3OnlineForHarkOptions &opts,
                                    kaldi::OnlineFeatureInterface *features_if);

  /// Returns the scaled log likelihood
  virtual kaldi::BaseFloat LogLikelihood(kaldi::int32 frame, kaldi::int32 index);

  virtual bool IsLastFrame(kaldi::int32 frame) const;

  virtual kaldi::int32 NumFramesReady() const;

  /// Indices are one-based!  This is for compatibility with OpenFst.
  virtual kaldi::int32 NumIndices() const { return trans_model_.NumTransitionIds(); }

 private:
  void ComputeForFrame(kaldi::int32 frame);

  kaldi::int32 NumSubsampledFrames(kaldi::int32) const;

  void DoNnetComputation(
      kaldi::int32 input_t_start,
      const kaldi::MatrixBase<kaldi::BaseFloat> &input_feats,
      const kaldi::VectorBase<kaldi::BaseFloat> &ivector,
      kaldi::int32 output_t_start,
      kaldi::int32 num_subsampled_frames);

  CachingOptimizingCompiler compiler_;

  kaldi::OnlineFeatureInterface *features_if_;
  const kaldi::nnet3::AmNnetSimple &nnet_;
  const kaldi::TransitionModel &trans_model_;
  DecodableNnet3OnlineForHarkOptions opts_;
  kaldi::CuVector<kaldi::BaseFloat> log_priors_;
  kaldi::int32 feat_dim_;
  kaldi::int32 left_context_;
  kaldi::int32 right_context_;
  kaldi::int32 num_pdfs_;
  kaldi::int32 begin_frame_;

  kaldi::Matrix<kaldi::BaseFloat> scaled_loglikes_;

  //KALDI_DISALLOW_COPY_AND_ASSIGN(DecodableNnet3SimpleOnlineForHark);
};

} // namespace nnet3
} // namespace kaldi

#endif // KALDI_DECODABLE_NNET_ONLINE_FOR_HARK_H_
