// recognizer.h

#ifdef _MSC_VER
#define NOMINMAX
#endif

#include <Poco/Format.h>
#include <Poco/StringTokenizer.h>
#include "recognizer.h"

namespace kaldi {
namespace nnet1 {

NnetRecognizer::NnetRecognizer(
        const kaldi::nnet1::Nnet& nnet,
        kaldi::nnet1::PdfPrior& pdf_prior,
        const kaldi::TransitionModel& trans_model,
        kaldi::nnet1::DecodableNnetOnlineForHarkOptions opt,
        Poco::SharedPtr<kaldi::OnlineFeatureMfcnet>& features,
        Poco::SharedPtr<kaldi::OnlineFeatureInterface>& features_if,
        const fst::Fst<fst::StdArc>* decode_fst,
        const kaldi::LatticeFasterDecoderConfig& decoder_config,
        const fst::SymbolTable* phone_syms,
        const fst::SymbolTable* word_syms,
        const kaldi::WordAlignLatticeLexiconInfo& lexicon_info,
        std::string& lm_name,
        int emulate_julius,
        Poco::Net::StreamSocket result_client):
    trans_model_(trans_model),
    features_(features), features_if_(features_if),
    decodable_(nnet, pdf_prior, trans_model, opt, features_if),
    decoder_(*decode_fst, decoder_config),
    phone_syms_(phone_syms), word_syms_(word_syms),
    lexicon_info_(lexicon_info),
    result_client_(result_client),
    lm_name_(lm_name),
    emulate_julius_(emulate_julius){
  // remove path prefix
#ifdef _MSC_VER
  std::string::size_type pos = lm_name.rfind("\\");
#else
  std::string::size_type pos = lm_name.rfind("/");
#endif
  lm_name_ = (pos != std::string::npos) ? lm_name.substr(++pos) : lm_name;
  emulate_julius_ = (emulate_julius != 0) ? 1 : 0;
  delimiter_sym_ = (emulate_julius_) ? ".\n" : "";
}

OnlineFeatureInterface* NnetRecognizer::GetOnlineFeatureInterface()
{
  return features_if_;
}

void NnetRecognizer::Recognize()
{
  if(!features_.isNull()){
    std::ostringstream str;
    str << std::fixed << std::setprecision(6)
        << "<SOURCEINFO SOURCEID=\"" << features_->GetSourceInfo().source_id
        <<           "\" AZIMUTH=\"" << features_->GetSourceInfo().azimuth
        <<         "\" ELEVATION=\"" << features_->GetSourceInfo().elevation
        <<               "\" SEC=\"" << features_->GetSourceInfo().tv_sec
        <<              "\" USEC=\"" << features_->GetSourceInfo().tv_usec << "\" />\n" << delimiter_sym_
        << "<STARTRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    try{
      result_client_.sendBytes(str.str().c_str(), str.str().size());
    }
    catch(Poco::Exception& e){
      KALDI_LOG << e.what();
    }
  }

  Timer timer; timer.Reset();
  //decoder_.Decode(&decodable_);
  decoder_.InitDecoding();
  for(;;){
    if(!features_.isNull()){
      features_->Receive();
    }
    decoder_.AdvanceDecoding(&decodable_);
    if(features_if_->IsLastFrame(features_if_->NumFramesReady() - 1)){
      decoder_.AdvanceDecoding(&decodable_); // <===ꂢ́H
      break;
    }
    if(decoder_.NumFramesDecoded() == features_if_->NumFramesReady()){
#ifdef _MSC_VER
      Sleep(1);
#else
      usleep(1000);
#endif
    }
  }
  decoder_.FinalizeDecoding();
  double elapsed = timer.Elapsed();
  KALDI_LOG << elapsed << " sec, RT=" << (elapsed * 100.0 / features_if_->NumFramesReady());

  if(!features_.isNull()){
    std::ostringstream str;
    str << "<ENDRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    try{
      result_client_.sendBytes(str.str().c_str(), str.str().size());
    }
    catch(Poco::Exception& e){
      KALDI_LOG << e.what();
    }
  }
}

void NnetRecognizer::WriteResult()
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0){
    std::ostringstream str;
    str << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
    try{
      result_client_.sendBytes(str.str().c_str(), str.str().size());
    }
    catch(Poco::Exception& e){
      KALDI_LOG << e.what();
    }
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned &weight);

  {
    std::ostringstream str;
    str << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
    if(lm_name_ != ""){ str << " NAME=\"" << lm_name_ << "\""; }
    str << std::fixed << std::setprecision(6)
        << ">\n"
        << " <SHYPO RANK=\"" << 1
        <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
        <<   "\" AMSCORE=\"" << -(double)weight.Value2()
        <<   "\" LMSCORE=\"" << -(double)weight.Value1() << "\">\n";
    size_t i = 0;
    //int word_index = 0;
    while(i < words.size()){
      std::string classid = word_syms_->Find(words[i]);
      std::string word;
      Poco::StringTokenizer tokenizer(classid, "+");
      word = tokenizer[0];
      //word = (word_index < tokenizer.count()) ? tokenizer[word_index] : tokenizer[0];
      ////switch(tokenizer.count()){
      ////case 2:
      ////  word = tokenizer[1];
      ////  break;
      ////case 1:
      ////case 4:
      ////default:
      ////  word = tokenizer[0];
      ////  break;
      ////}
      if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      str << std::fixed << std::setprecision(3)
          << "  <WHYPO WORD=\"" << word
          <<    "\" CLASSID=\"" << classid
          <<      "\" PHONE=\"" << ""
          <<         "\" CM=\"" << 1.0f << "\"/>\n";
      i++;
    }
    str << " </SHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

    try{
      result_client_.sendBytes(str.str().c_str(), str.str().size());
    }
    catch(Poco::Exception& e){
      KALDI_LOG << e.what();
    }
  }
}

void NnetRecognizer::PrintResult(std::ostream& out)
{
  std::stringstream ss;
  if(decoder_.NumFramesDecoded() <= 0){
    if(emulate_julius_){
      ss << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n"
         << "### Recognition: 2nd pass (RL heuristic best-first) \n"
         << "STAT: 00 \n"
         << "sentence1: \n"
         << "wseq1: \n"
         << "phseq1: \n"
         << "cmscore1: \n"
         << "score1: \n\n"
         << "ERROR: recognition failed.";
      out << ss.str() << std::endl;
    }
    else{
      ss << "recognition failed\n"
         << "result: \n"
         << "phone_seq: ";
      out << ss.str() << std::endl;
    }
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned, &weight);

  if(emulate_julius_){
    if(!features_.isNull()){
      ss << std::fixed << std::setprecision(6)
         <<   "source_id = " << features_->GetSourceInfo().source_id
         <<   ", azimuth = " << features_->GetSourceInfo().azimuth
         << ", elevation = " << features_->GetSourceInfo().elevation
         <<       ", sec = " << features_->GetSourceInfo().tv_sec
         <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
    }
    ss << "### Recognition: 2nd pass (RL heuristic best-first) \n"
       << "STAT: 00 \n"
       << "sentence1: ";
    if(word_syms_ != NULL){
      size_t i = 0;
      //int word_index = 0;
      while(i < words.size()){
        std::string classid = word_syms_->Find(words[i]);
        Poco::StringTokenizer tokenizer(classid, "+");
        //ss << (word_index < tokenizer.count()) ? tokenizer[word_index] : tokenizer[0];
        ss << tokenizer[0] << " ";
        if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
        i++;
      }
    }
    else{
      size_t i = 0;
      while(i < words.size()){ ss << words[i++] << " "; }
    }
    ss << "\n";
  }

  ss << (emulate_julius_ ? "wseq1: " : "result: ");
  if(word_syms_ != NULL){
    size_t i = 0;
    while(i < words.size()){
      std::string s = word_syms_->Find(words[i]);
      if(s == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss << s << " ";
      i++;
    }
  }
  else{
    size_t i = 0;
    while(i < words.size()){ ss << words[i++] << " "; }
  }
  ss << "\n";

  ss << (emulate_julius_ ? "phseq1: " : "phone_seq: ");
  if(phone_syms_ != NULL){
    size_t i = 0;
    while(i < phones_aligned.size()){
      kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(phones_aligned[i]);
      kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
      std::string s = phone_syms_->Find(phone);
      if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
      ss << s << " ";
      i++;
    }
  }
  ss << "\n";

  if(emulate_julius_){
    ss << "cmscore1: ";
    size_t i = 0;
    while(i < words.size()){
      ss << std::fixed << std::setprecision(3) << 1.0f << " ";
      i++;
    }
    ss << "\n";
    ss << std::fixed << std::setprecision(6)
       << "score1: " << -(double)(weight.Value1() + weight.Value2())
       << " ( AM: " << -(double)weight.Value2() << ", LM: " << -(double)weight.Value1() << " )";
  }
  else{
    ss << std::fixed << std::setprecision(6)
       <<     "score=" << -(double)(weight.Value1() + weight.Value2())
       << ", amscore=" << -(double)weight.Value2()
       << ", lmscore=" << -(double)weight.Value1();
  }
  out << ss.str() << std::endl;

  /*
  Lattice lat;
  decoder.GetRawLattice(&lat);
  CompactLattice clat;
  if(!DeterminizeLatticePhonePrunedWrapper(trans_model, &lat, decoder_config.lattice_beam, &clat, decoder_config.det_opts)){ }
  CompactLattice best_path_clat;
  CompactLatticeShortestPath(clat, &best_path_clat);

  Lattice best_path_lat;
  ConvertLattice(best_path_clat, &best_path_lat);
  */
}
} // namespace nnet1


namespace nnet2 {
NnetRecognizer::NnetRecognizer(
        const kaldi::nnet1::Nnet& nnet,
        kaldi::nnet1::PdfPrior& pdf_prior,
        const kaldi::TransitionModel& trans_model,
        kaldi::nnet1::DecodableNnetOnlineForHarkOptions opt,
        Poco::SharedPtr<kaldi::OnlineFeatureMfcnet>& features,
        Poco::SharedPtr<kaldi::OnlineFeatureInterface>& features_if,
        const fst::Fst<fst::StdArc>* decode_fst,
        const kaldi::LatticeFasterDecoderConfig& decoder_config,
        const fst::SymbolTable* phone_syms,
        const fst::SymbolTable* word_syms,
        const kaldi::WordAlignLatticeLexiconInfo& lexicon_info,
        std::string& lm_name,
        int emulate_julius,
        Poco::Net::StreamSocket result_client):
    trans_model_(trans_model),
    features_(features), features_if_(features_if),
    decodable_(nnet, pdf_prior, trans_model, opt, features_if),
    decoder_(*decode_fst, decoder_config),
    phone_syms_(phone_syms), word_syms_(word_syms),
    lexicon_info_(lexicon_info),
    result_client_(result_client),
    lm_name_(lm_name),
    emulate_julius_(emulate_julius){
  // remove path prefix
#ifdef _MSC_VER
  std::string::size_type pos = lm_name.rfind("\\");
#else
  std::string::size_type pos = lm_name.rfind("/");
#endif
  lm_name_ = (pos != std::string::npos) ? lm_name.substr(++pos) : lm_name;
  emulate_julius_ = (emulate_julius != 0) ? 1 : 0;
  delimiter_sym_ = (emulate_julius_) ? ".\n" : "";
}

OnlineFeatureInterface* NnetRecognizer::GetOnlineFeatureInterface()
{
  return features_if_;
}

void NnetRecognizer::Recognize()
{
  if(!features_.isNull()){
    std::ostringstream str;
    str << std::fixed << std::setprecision(6)
        << "<SOURCEINFO SOURCEID=\"" << features_->GetSourceInfo().source_id
        <<           "\" AZIMUTH=\"" << features_->GetSourceInfo().azimuth
        <<         "\" ELEVATION=\"" << features_->GetSourceInfo().elevation
        <<               "\" SEC=\"" << features_->GetSourceInfo().tv_sec
        <<              "\" USEC=\"" << features_->GetSourceInfo().tv_usec << "\" />\n" << delimiter_sym_
        << "<STARTRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    try{
      result_client_.sendBytes(str.str().c_str(), str.str().size());
    }
    catch(Poco::Exception& e){
      KALDI_LOG << e.what();
    }
  }

  Timer timer; timer.Reset();
  //decoder_.Decode(&decodable_);
  decoder_.InitDecoding();
  for(;;){
    if(!features_.isNull()){
      features_->Receive();
    }
    decoder_.AdvanceDecoding(&decodable_);
    if(features_if_->IsLastFrame(features_if_->NumFramesReady() - 1)){
      decoder_.AdvanceDecoding(&decodable_); // <===ꂢ́H
      break;
    }
    if(decoder_.NumFramesDecoded() == features_if_->NumFramesReady()){
#ifdef _MSC_VER
      Sleep(1);
#else
      usleep(1000);
#endif
    }
  }
  decoder_.FinalizeDecoding();
  double elapsed = timer.Elapsed();
  KALDI_LOG << elapsed << " sec, RT=" << (elapsed * 100.0 / features_if_->NumFramesReady());

  if(!features_.isNull()){
    std::ostringstream str;
    str << "<ENDRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    try{
      result_client_.sendBytes(str.str().c_str(), str.str().size());
    }
    catch(Poco::Exception& e){
      KALDI_LOG << e.what();
    }
  }
}

void NnetRecognizer::WriteResult()
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0){
    std::ostringstream str;
    str << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
    try{
      result_client_.sendBytes(str.str().c_str(), str.str().size());
    }
    catch(Poco::Exception& e){
      KALDI_LOG << e.what();
    }
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned &weight);

  {
    std::ostringstream str;
    str << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
    if(lm_name_ != ""){ str << " NAME=\"" << lm_name_ << "\""; }
    str << std::fixed << std::setprecision(6)
        << ">\n"
        << " <SHYPO RANK=\"" << 1
        <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
        <<   "\" AMSCORE=\"" << -(double)weight.Value2()
        <<   "\" LMSCORE=\"" << -(double)weight.Value1() << "\">\n";
    size_t i = 0;
    //int word_index = 0;
    while(i < words.size()){
      std::string classid = word_syms_->Find(words[i]);
      std::string word;
      Poco::StringTokenizer tokenizer(classid, "+");
      word = tokenizer[0];
      //word = (word_index < tokenizer.count()) ? tokenizer[word_index] : tokenizer[0];
      ////switch(tokenizer.count()){
      ////case 2:
      ////  word = tokenizer[1];
      ////  break;
      ////case 1:
      ////case 4:
      ////default:
      ////  word = tokenizer[0];
      ////  break;
      ////}
      if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      str << std::fixed << std::setprecision(3)
          << "  <WHYPO WORD=\"" << word
          <<    "\" CLASSID=\"" << classid
          <<      "\" PHONE=\"" << ""
          <<         "\" CM=\"" << 1.0f << "\"/>\n";
      i++;
    }
    str << " </SHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

    try{
      result_client_.sendBytes(str.str().c_str(), str.str().size());
    }
    catch(Poco::Exception& e){
      KALDI_LOG << e.what();
    }
  }
}

void NnetRecognizer::PrintResult(std::ostream& out)
{
  std::stringstream ss;
  if(decoder_.NumFramesDecoded() <= 0){
    if(emulate_julius_){
      ss << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n"
         << "### Recognition: 2nd pass (RL heuristic best-first) \n"
         << "STAT: 00 \n"
         << "sentence1: \n"
         << "wseq1: \n"
         << "phseq1: \n"
         << "cmscore1: \n"
         << "score1: \n\n"
         << "ERROR: recognition failed.";
      out << ss.str() << std::endl;
    }
    else{
      ss << "recognition failed\n"
         << "result: \n"
         << "phone_seq: ";
      out << ss.str() << std::endl;
    }
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned, &weight);

  if(emulate_julius_){
    ss << std::fixed << std::setprecision(6)
       <<   "source_id = " << features_->GetSourceInfo().source_id
       <<   ", azimuth = " << features_->GetSourceInfo().azimuth
       << ", elevation = " << features_->GetSourceInfo().elevation
       <<       ", sec = " << features_->GetSourceInfo().tv_sec
       <<      ", usec = " << features_->GetSourceInfo().tv_usec
       << "### Recognition: 2nd pass (RL heuristic best-first) \n"
       << "STAT: 00 \n"
       << "sentence1: ";
    if(word_syms_ != NULL){
      size_t i = 0;
      //int word_index = 0;
      while(i < words.size()){
        std::string classid = word_syms_->Find(words[i]);
        Poco::StringTokenizer tokenizer(classid, "+");
        //ss << (word_index < tokenizer.count()) ? tokenizer[word_index] : tokenizer[0];
        ss << tokenizer[0] << " ";
        if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
        i++;
      }
    }
    else{
      size_t i = 0;
      while(i < words.size()){ ss << words[i++] << " "; }
    }
    ss << "\n";
  }

  ss << (emulate_julius_) ? "wseq1: " : "result: ";
  if(word_syms_ != NULL){
    size_t i = 0;
    while(i < words.size()){
      std::string s = word_syms_->Find(words[i]);
      if(s == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss << s << " ";
      i++;
    }
    ss << "\n";
  }
  else{
    size_t i = 0;
    while(i < words.size()){ ss << words[i++] << " "; }
    ss << "\n";
  }

  ss << (emulate_julius_) ? "phseq1: " : "phone_seq: ";
  if(phone_syms_ != NULL){
    size_t i = 0;
    while(i < phones_aligned.size()){
      kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(phones_aligned[i]);
      kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
      std::string s = phone_syms_->Find(phone);
      if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
      ss << s << " ";
      i++;
    }
    ss << "\n";
  }

  if(emulate_julius_){
    ss << "cmscore1: ";
    size_t i = 0;
    while(i < words.size()){
      ss << std::fixed << std::setprecision(3) << 1.0f << " ";
      i++;
    }
    ss << "\n";
    ss << std::fixed << std::setprecision(6)
       << "score1: " << -(double)(weight.Value1() + weight.Value2())
       << " ( AM: " << -(double)weight.Value2() << ", LM: " << -(double)weight.Value1() << " )";
  }
  else{
    ss << std::fixed << std::setprecision(6)
       <<     "score=" << -(double)(weight.Value1() + weight.Value2())
       << ", amscore=" << -(double)weight.Value2()
       << ", lmscore=" << -(double)weight.Value1();
  }
  out << ss.str() << std::endl;

  /*
  Lattice lat;
  decoder.GetRawLattice(&lat);
  CompactLattice clat;
  if(!DeterminizeLatticePhonePrunedWrapper(trans_model, &lat, decoder_config.lattice_beam, &clat, decoder_config.det_opts)){ }
  CompactLattice best_path_clat;
  CompactLatticeShortestPath(clat, &best_path_clat);

  Lattice best_path_lat;
  ConvertLattice(best_path_clat, &best_path_lat);
  */
}
} // namespace nnet2


namespace nnet3 {
NnetRecognizer::NnetRecognizer(
        const kaldi::nnet1::Nnet& nnet,
        kaldi::nnet1::PdfPrior& pdf_prior,
        const kaldi::TransitionModel& trans_model,
        kaldi::nnet1::DecodableNnetOnlineForHarkOptions opt,
        Poco::SharedPtr<kaldi::OnlineFeatureMfcnet>& features,
        Poco::SharedPtr<kaldi::OnlineFeatureInterface>& features_if,
        const fst::Fst<fst::StdArc>* decode_fst,
        const kaldi::LatticeFasterDecoderConfig& decoder_config,
        const fst::SymbolTable* phone_syms,
        const fst::SymbolTable* word_syms,
        const kaldi::WordAlignLatticeLexiconInfo& lexicon_info,
        std::string& lm_name,
        int emulate_julius,
        Poco::Net::StreamSocket result_client):
    trans_model_(trans_model),
    features_(features), features_if_(features_if),
    decodable_(nnet, pdf_prior, trans_model, opt, features_if),
    decoder_(*decode_fst, decoder_config),
    phone_syms_(phone_syms), word_syms_(word_syms),
    lexicon_info_(lexicon_info),
    result_client_(result_client),
    lm_name_(lm_name),
    emulate_julius_(emulate_julius){
  // remove path prefix
#ifdef _MSC_VER
  std::string::size_type pos = lm_name.rfind("\\");
#else
  std::string::size_type pos = lm_name.rfind("/");
#endif
  lm_name_ = (pos != std::string::npos) ? lm_name.substr(++pos) : lm_name;
  emulate_julius_ = (emulate_julius != 0) ? 1 : 0;
  delimiter_sym_ = (emulate_julius_) ? ".\n" : "";
}

OnlineFeatureInterface* NnetRecognizer::GetOnlineFeatureInterface()
{
  return features_if_;
}

void NnetRecognizer::Recognize()
{
  if(!features_.isNull()){
    std::ostringstream str;
    str << std::fixed << std::setprecision(6)
        << "<SOURCEINFO SOURCEID=\"" << features_->GetSourceInfo().source_id
        <<           "\" AZIMUTH=\"" << features_->GetSourceInfo().azimuth
        <<         "\" ELEVATION=\"" << features_->GetSourceInfo().elevation
        <<               "\" SEC=\"" << features_->GetSourceInfo().tv_sec
        <<              "\" USEC=\"" << features_->GetSourceInfo().tv_usec << "\" />\n" << delimiter_sym_
        << "<STARTRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    try{
      result_client_.sendBytes(str.str().c_str(), str.str().size());
    }
    catch(Poco::Exception& e){
      KALDI_LOG << e.what();
    }
  }

  Timer timer; timer.Reset();
  //decoder_.Decode(&decodable_);
  decoder_.InitDecoding();
  for(;;){
    if(!features_.isNull()){
      features_->Receive();
    }
    decoder_.AdvanceDecoding(&decodable_);
    if(features_if_->IsLastFrame(features_if_->NumFramesReady() - 1)){
      decoder_.AdvanceDecoding(&decodable_); // <===ꂢ́H
      break;
    }
    if(decoder_.NumFramesDecoded() == features_if_->NumFramesReady()){
#ifdef _MSC_VER
      Sleep(1);
#else
      usleep(1000);
#endif
    }
  }
  decoder_.FinalizeDecoding();
  double elapsed = timer.Elapsed();
  KALDI_LOG << elapsed << " sec, RT=" << (elapsed * 100.0 / features_if_->NumFramesReady());

  if(!features_.isNull()){
    std::ostringstream str;
    str << "<ENDRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    try{
      result_client_.sendBytes(str.str().c_str(), str.str().size());
    }
    catch(Poco::Exception& e){
      KALDI_LOG << e.what();
    }
  }
}

void NnetRecognizer::WriteResult()
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0){
    std::ostringstream str;
    str << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
    try{
      result_client_.sendBytes(str.str().c_str(), str.str().size());
    }
    catch(Poco::Exception& e){
      KALDI_LOG << e.what();
    }
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned &weight);

  {
    std::ostringstream str;
    str << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
    if(lm_name_ != ""){ str << " NAME=\"" << lm_name_ << "\""; }
    str << std::fixed << std::setprecision(6)
        << ">\n"
        << " <SHYPO RANK=\"" << 1
        <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
        <<   "\" AMSCORE=\"" << -(double)weight.Value2()
        <<   "\" LMSCORE=\"" << -(double)weight.Value1() << "\">\n";
    size_t i = 0;
    //int word_index = 0;
    while(i < words.size()){
      std::string classid = word_syms_->Find(words[i]);
      std::string word;
      Poco::StringTokenizer tokenizer(classid, "+");
      word = tokenizer[0];
      //word = (word_index < tokenizer.count()) ? tokenizer[word_index] : tokenizer[0];
      ////switch(tokenizer.count()){
      ////case 2:
      ////  word = tokenizer[1];
      ////  break;
      ////case 1:
      ////case 4:
      ////default:
      ////  word = tokenizer[0];
      ////  break;
      ////}
      if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      str << std::fixed << std::setprecision(3)
          << "  <WHYPO WORD=\"" << word
          <<    "\" CLASSID=\"" << classid
          <<      "\" PHONE=\"" << ""
          <<         "\" CM=\"" << 1.0f << "\"/>\n";
      i++;
    }
    str << " </SHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

    try{
      result_client_.sendBytes(str.str().c_str(), str.str().size());
    }
    catch(Poco::Exception& e){
      KALDI_LOG << e.what();
    }
  }
}

void NnetRecognizer::PrintResult(std::ostream& out)
{
  std::stringstream ss;
  if(decoder_.NumFramesDecoded() <= 0){
    if(emulate_julius_){
      ss << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n"
         << "### Recognition: 2nd pass (RL heuristic best-first) \n"
         << "STAT: 00 \n"
         << "sentence1: \n"
         << "wseq1: \n"
         << "phseq1: \n"
         << "cmscore1: \n"
         << "score1: \n\n"
         << "ERROR: recognition failed.";
      out << ss.str() << std::endl;
    }
    else{
      ss << "recognition failed\n"
         << "result: \n"
         << "phone_seq: ";
      out << ss.str() << std::endl;
    }
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned, &weight);

  if(emulate_julius_){
    ss << std::fixed << std::setprecision(6)
       <<   "source_id = " << features_->GetSourceInfo().source_id
       <<   ", azimuth = " << features_->GetSourceInfo().azimuth
       << ", elevation = " << features_->GetSourceInfo().elevation
       <<       ", sec = " << features_->GetSourceInfo().tv_sec
       <<      ", usec = " << features_->GetSourceInfo().tv_usec
       << "### Recognition: 2nd pass (RL heuristic best-first) \n"
       << "STAT: 00 \n"
       << "sentence1: ";
    if(word_syms_ != NULL){
      size_t i = 0;
      //int word_index = 0;
      while(i < words.size()){
        std::string classid = word_syms_->Find(words[i]);
        Poco::StringTokenizer tokenizer(classid, "+");
        //ss << (word_index < tokenizer.count()) ? tokenizer[word_index] : tokenizer[0];
        ss << tokenizer[0] << " ";
        if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
        i++;
      }
    }
    else{
      size_t i = 0;
      while(i < words.size()){ ss << words[i++] << " "; }
    }
    ss << "\n";
  }

  ss << (emulate_julius_) ? "wseq1: " : "result: ";
  if(word_syms_ != NULL){
    size_t i = 0;
    while(i < words.size()){
      std::string s = word_syms_->Find(words[i]);
      if(s == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss << s << " ";
      i++;
    }
    ss << "\n";
  }
  else{
    size_t i = 0;
    while(i < words.size()){ ss << words[i++] << " "; }
    ss << "\n";
  }

  ss << (emulate_julius_) ? "phseq1: " : "phone_seq: ";
  if(phone_syms_ != NULL){
    size_t i = 0;
    while(i < phones_aligned.size()){
      kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(phones_aligned[i]);
      kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
      std::string s = phone_syms_->Find(phone);
      if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
      ss << s << " ";
      i++;
    }
    ss << "\n";
  }

  if(emulate_julius_){
    ss << "cmscore1: ";
    size_t i = 0;
    while(i < words.size()){
      ss << std::fixed << std::setprecision(3) << 1.0f << " ";
      i++;
    }
    ss << "\n";
    ss << std::fixed << std::setprecision(6)
       << "score1: " << -(double)(weight.Value1() + weight.Value2())
       << " ( AM: " << -(double)weight.Value2() << ", LM: " << -(double)weight.Value1() << " )";
  }
  else{
    ss << std::fixed << std::setprecision(6)
       <<     "score=" << -(double)(weight.Value1() + weight.Value2())
       << ", amscore=" << -(double)weight.Value2()
       << ", lmscore=" << -(double)weight.Value1();
  }
  out << ss.str() << std::endl;

  /*
  Lattice lat;
  decoder.GetRawLattice(&lat);
  CompactLattice clat;
  if(!DeterminizeLatticePhonePrunedWrapper(trans_model, &lat, decoder_config.lattice_beam, &clat, decoder_config.det_opts)){ }
  CompactLattice best_path_clat;
  CompactLatticeShortestPath(clat, &best_path_clat);

  Lattice best_path_lat;
  ConvertLattice(best_path_clat, &best_path_lat);
  */
}
} // namespace nnet3
} // namespace kaldi

